import binascii
import enum
import gzip
import io
import json
import logging
import mimetypes
import os
import warnings
from collections.abc import Iterator
from typing import Any
from typing import AnyStr
from typing import IO
from urllib.parse import quote

from gcloud.aio.auth import AioSession  # pylint: disable=no-name-in-module
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.auth import Token  # pylint: disable=no-name-in-module

from .bucket import Bucket
from .constants import DEFAULT_TIMEOUT

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from time import sleep
    from requests import HTTPError as ResponseError
    from requests import Session
    from builtins import open as file_open
else:
    from aiofiles import open as file_open  # type: ignore[no-redef]
    from asyncio import sleep  # type: ignore[assignment]
    from aiohttp import (  # type: ignore[assignment]
        ClientResponseError as ResponseError,
    )
    from aiohttp import ClientSession as Session  # type: ignore[assignment]

MAX_CONTENT_LENGTH_SIMPLE_UPLOAD = 5 * 1024 * 1024  # 5 MB
SCOPES = [
    'https://www.googleapis.com/auth/devstorage.full_control',
]

log = logging.getLogger(__name__)


def init_api_root(api_root: str | None) -> tuple[bool, str]:
    if api_root:
        return True, api_root

    host = os.environ.get('STORAGE_EMULATOR_HOST')
    if host:
        if not host.startswith('http'):
            warnings.warn('STORAGE_EMULATOR_HOST must include http:// prefix',
                          DeprecationWarning)
            host = f'http://{host}'
        return True, host

    return False, 'https://www.googleapis.com'


def choose_boundary() -> str:
    """Stolen from urllib3.filepost.choose_boundary() as of v1.26.2."""
    return binascii.hexlify(os.urandom(16)).decode('ascii')


def encode_multipart_formdata(
    fields: list[tuple[dict[str, str], bytes]],
    boundary: str,
) -> tuple[bytes, str]:
    """
    Stolen from urllib3.filepost.encode_multipart_formdata() as of v1.26.2.

    Very heavily modified to be compatible with our gcloud-rest converter and
    to avoid unnecessary urllib3 dependencies (since that's only included with
    requests, not aiohttp).
    """
    body: list[bytes] = []
    for headers, data in fields:
        body.append(f'--{boundary}\r\n'.encode())

        # The below is from RequestFields.render_headers()
        # Since we only use Content-Type, we could simplify the below to a
        # single line... but probably best to be safe for future modifications.
        for field in [
            'Content-Disposition', 'Content-Type',
            'Content-Location',
        ]:
            value = headers.pop(field, None)
            if value:
                body.append(f'{field}: {value}\r\n'.encode())
        for field, value in headers.items():
            # N.B. potential bug copied from urllib3 code; zero values should
            # be sent! Keeping it for now, since Google libs use urllib3 for
            # their examples.
            if value:
                body.append(f'{field}: {value}\r\n'.encode())

        body.append(b'\r\n')
        body.append(data)
        body.append(b'\r\n')

    body.append(f'--{boundary}--\r\n'.encode())

    # N.B. 'multipart/form-data' in upstream, but Google wants 'related'
    content_type = f'multipart/related; boundary={boundary}'

    return b''.join(body), content_type


class UploadType(enum.Enum):
    SIMPLE = 1
    RESUMABLE = 2
    MULTIPART = 3  # unused: SIMPLE upgrades to MULTIPART when metadata exists


class StreamResponse:
    """
    This class provides an abstraction between the slightly different
    recommended streaming implementations between requests and aiohttp.
    """

    def __init__(self, response: Any) -> None:
        self._response = response
        self._iter: Iterator[bytes] | None = None

    @property
    def content_length(self) -> int:
        return int(self._response.headers.get('content-length', 0))

    async def read(self, size: int = -1) -> bytes:
        chunk: bytes
        if BUILD_GCLOUD_REST:
            if self._iter is None:
                self._iter = self._response.iter_content(chunk_size=size)
            chunk = next(self._iter, b'')
        else:
            chunk = await self._response.content.read(size)
        return chunk

    async def __aenter__(self) -> Any:
        # strictly speaking, since this method can't be called via gcloud-rest,
        # we know the return type is aiohttp.ClientResponse
        return await self._response.__aenter__()

    async def __aexit__(self, *exc_info: Any) -> None:
        await self._response.__aexit__(*exc_info)


class Storage:
    _api_root: str
    _api_is_dev: bool
    _api_root_read: str
    _api_root_write: str

    def __init__(
            self, *, service_file: str | IO[AnyStr] | None = None,
            token: Token | None = None, session: Session | None = None,
            api_root: str | None = None,
    ) -> None:
        self._api_is_dev, self._api_root = init_api_root(api_root)
        self._api_root_read = f'{self._api_root}/storage/v1/b'
        self._api_root_write = f'{self._api_root}/upload/storage/v1/b'

        self.session = AioSession(session, verify_ssl=not self._api_is_dev)
        self.token = token or Token(
            service_file=service_file, scopes=SCOPES,
            session=self.session.session,  # type: ignore[arg-type]
        )

    async def _headers(self) -> dict[str, str]:
        if self._api_is_dev:
            return {}

        token = await self.token.get()
        return {
            'Authorization': f'Bearer {token}',
        }

    # This method makes the following API call:
    # https://cloud.google.com/storage/docs/json_api/v1/buckets/list
    async def list_buckets(
        self, project: str, *,
        params: dict[str, str] | None = None,
        headers: dict[str, Any] | None = None,
        session: Session | None = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> list[Bucket]:
        url = f'{self._api_root_read}?project={project}'
        headers = headers or {}
        headers.update(await self._headers())
        params = params or {}
        if not params.get('pageToken'):
            params['pageToken'] = ''
        s = AioSession(session) if session else self.session
        buckets = []

        while True:
            resp = await s.get(url, headers=headers,
                               params=params or {},
                               timeout=timeout)

            content: dict[str, Any] = await resp.json(content_type=None)
            for item in content.get('items', []):
                buckets.append(Bucket(self, item['id']))

            params['pageToken'] = content.get('nextPageToken', '')
            if not params['pageToken']:
                break
        return buckets

    def get_bucket(self, bucket_name: str) -> Bucket:
        return Bucket(self, bucket_name)

    async def copy(
        self, bucket: str, object_name: str,
        destination_bucket: str, *, new_name: str | None = None,
        metadata: dict[str, Any] | None = None,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> dict[str, Any]:
        """
        When files are too large, multiple calls to ``rewriteTo`` are made. We
        refer to the same copy job by using the ``rewriteToken`` from the
        previous return payload in subsequent ``rewriteTo`` calls.

        Using the ``rewriteTo`` GCS API is preferred in part because it is able
        to make multiple calls to fully copy an object whereas the ``copyTo``
        GCS API only calls ``rewriteTo`` once under the hood, and thus may fail
        if files are large.

        In the rare case you need to resume a copy operation, include the
        ``rewriteToken`` in the ``params`` dictionary. Once you begin a
        multi-part copy operation, you then have 1 week to complete the copy
        job.

        See https://cloud.google.com/storage/docs/json_api/v1/objects/rewrite
        """
        # pylint: disable=too-many-locals
        if not new_name:
            new_name = object_name

        url = (
            f'{self._api_root_read}/{bucket}/o/'
            f'{quote(object_name, safe="")}/rewriteTo/b/'
            f'{destination_bucket}/o/{quote(new_name, safe="")}'
        )

        # We may optionally supply metadata* to apply to the rewritten
        # object, which explains why `rewriteTo` is a POST endpoint; when no
        # metadata is given, we have to send an empty body.
        # * https://cloud.google.com/storage/docs/json_api/v1/objects#resource
        metadict = (metadata or {}).copy()
        metadict = {
            self._format_metadata_key(k): v
            for k, v in metadict.items()
        }
        if 'metadata' in metadict:
            metadict['metadata'] = {
                str(k): str(v) if v is not None else None
                for k, v in metadict['metadata'].items()
            }

        metadata_ = json.dumps(metadict)

        headers = headers or {}
        headers.update(await self._headers())
        headers.update({
            'Content-Length': str(len(metadata_)),
            'Content-Type': 'application/json; charset=UTF-8',
        })

        params = params or {}

        s = AioSession(session) if session else self.session
        resp = await s.post(
            url, headers=headers, params=params, timeout=timeout,
            data=metadata_,
        )

        data: dict[str, Any] = await resp.json(content_type=None)

        while not data.get('done') and data.get('rewriteToken'):
            params['rewriteToken'] = data['rewriteToken']
            resp = await s.post(
                url, headers=headers, params=params,
                timeout=timeout, data=metadata_,
            )
            data = await resp.json(content_type=None)

        return data

    async def delete(
        self, bucket: str, object_name: str, *,
        timeout: int = DEFAULT_TIMEOUT,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        session: Session | None = None,
    ) -> str:
        # https://cloud.google.com/storage/docs/request-endpoints#encoding
        encoded_object_name = quote(object_name, safe='')
        url = f'{self._api_root_read}/{bucket}/o/{encoded_object_name}'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        resp = await s.delete(
            url, headers=headers, params=params or {},
            timeout=timeout,
        )

        try:
            data: str = await resp.text()
        except (AttributeError, TypeError):
            data = str(resp.text)

        return data

    async def download(
        self, bucket: str, object_name: str, *,
        headers: dict[str, Any] | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> bytes:
        return await self._download(
            bucket, object_name, headers=headers,
            timeout=timeout, params={'alt': 'media'},
            session=session,
        )

    async def download_to_filename(
        self, bucket: str, object_name: str,
        filename: str, **kwargs: Any,
    ) -> None:
        async with file_open(  # type: ignore[attr-defined]
                filename,
                mode='wb+',
        ) as file_object:
            await file_object.write(
                await self.download(bucket, object_name, **kwargs),
            )

    async def download_metadata(
        self, bucket: str, object_name: str, *,
        headers: dict[str, Any] | None = None,
        session: Session | None = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> dict[str, Any]:
        data = await self._download(
            bucket, object_name, headers=headers,
            timeout=timeout, params={'alt': 'json'},
            session=session,
        )
        metadata: dict[str, Any] = json.loads(data.decode())
        return metadata

    async def download_stream(
        self, bucket: str, object_name: str, *,
        headers: dict[str, Any] | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> StreamResponse:
        """
        Download a GCS object in a buffered stream.

        Args:
            bucket: The bucket from which to download.
            object_name: The object within the bucket to download.
            headers: Custom header values for the request, such as range.
            timeout: Timeout, in seconds, for the request. Note that with this
                function, this is the time to the beginning of the response
                data (TTFB).
            session: A specific session to (re)use.

        Returns:
            StreamResponse: A object encapsulating the stream, similar to
            io.BufferedIOBase, but it only supports the read() function.
        """
        return await self._download_stream(
            bucket, object_name,
            headers=headers, timeout=timeout,
            params={'alt': 'media'},
            session=session,
        )

    async def list_objects(
        self, bucket: str, *,
        params: dict[str, str] | None = None,
        headers: dict[str, Any] | None = None,
        session: Session | None = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> dict[str, Any]:
        url = f'{self._api_root_read}/{bucket}/o'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        resp = await s.get(
            url, headers=headers, params=params or {},
            timeout=timeout,
        )
        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    # https://cloud.google.com/storage/docs/json_api/v1/how-tos/upload
    # pylint: disable=too-many-locals
    async def upload(
        self, bucket: str, object_name: str, file_data: Any,
        *, content_type: str | None = None,
        parameters: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        metadata: dict[str, Any] | None = None,
        session: Session | None = None,
        force_resumable_upload: bool | None = None,
        zipped: bool = False,
        timeout: int = 30,
    ) -> dict[str, Any]:
        url = f'{self._api_root_write}/{bucket}/o'
        stream = self._preprocess_data(file_data)

        parameters = parameters or {}
        if zipped:
            parameters['contentEncoding'] = 'gzip'
            # Here we load the file-like object data into memory in chunks and
            # re-write it compressed. This is implemented like this so we don't
            # load the whole file into memory at once.
            stream = self._compress_file_in_chunks(input_stream=stream)

        if BUILD_GCLOUD_REST and isinstance(stream, io.StringIO):
            # HACK: `requests` library does not accept `str` as `data` in `put`
            # HTTP request.
            stream = io.BytesIO(stream.getvalue().encode('utf-8'))

        content_length = self._get_stream_len(stream)

        # mime detection method same as in aiohttp 3.4.4
        content_type = content_type or mimetypes.guess_type(object_name)[0]

        headers = headers or {}
        headers.update(await self._headers())
        headers.update({
            'Content-Length': str(content_length),
            'Content-Type': content_type or '',
        })

        upload_type = self._decide_upload_type(
            force_resumable_upload,
            content_length,
        )
        log.debug('using %r gcloud storage upload method', upload_type)

        if upload_type == UploadType.RESUMABLE:
            return await self._upload_resumable(
                url, object_name, stream, parameters, headers,
                metadata=metadata, session=session, timeout=timeout,
            )
        if upload_type == UploadType.SIMPLE:
            if metadata:
                return await self._upload_multipart(
                    url, object_name, stream, parameters, headers, metadata,
                    session=session, timeout=timeout,
                )
            return await self._upload_simple(
                url, object_name, stream, parameters, headers, session=session,
                timeout=timeout,
            )

        raise TypeError(f'upload type {upload_type} not supported')

    async def upload_from_filename(
        self, bucket: str, object_name: str,
        filename: str,
        **kwargs: Any,
    ) -> dict[str, Any]:
        async with file_open(  # type: ignore[attr-defined]
                filename,
                mode='rb',
        ) as file_object:
            contents = await file_object.read()
            return await self.upload(bucket, object_name, contents, **kwargs)

    # https://cloud.google.com/storage/docs/json_api/v1/objects/compose
    async def compose(
        self, bucket: str, object_name: str,
        source_object_names: list[str], *,
        content_type: str | None = None,
        params: dict[str, str] | None = None,
        headers: dict[str, Any] | None = None,
        session: Session | None = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> dict[str, Any]:
        url = (
            f'{self._api_root_read}/{bucket}/o/'
            f'{quote(object_name, safe="")}/compose'
        )
        headers = headers or {}
        headers.update(await self._headers())
        params = params or {}

        payload: dict[str, Any] = {
            'sourceObjects': [{'name': name} for name in source_object_names],
        }
        if content_type:
            payload['destination'] = {'contentType': content_type}
        body = json.dumps(payload).encode('utf-8')
        headers.update({
            'Content-Length': str(len(body)),
            'Content-Type': 'application/json; charset=UTF-8',
        })

        s = AioSession(session) if session else self.session
        resp = await s.post(
            url, headers=headers, params=params, timeout=timeout,
            data=body,
        )
        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    @staticmethod
    def _get_stream_len(stream: IO[AnyStr]) -> int:
        current = stream.tell()
        try:
            return stream.seek(0, os.SEEK_END)
        finally:
            stream.seek(current)

    @staticmethod
    def _preprocess_data(data: Any) -> IO[Any]:
        if data is None:
            return io.StringIO('')

        if isinstance(data, bytes):
            return io.BytesIO(data)
        if isinstance(data, str):
            return io.StringIO(data)
        if isinstance(data, io.IOBase):
            return data  # type: ignore[return-value]

        raise TypeError(f'unsupported upload type: "{type(data)}"')

    @staticmethod
    def _compress_file_in_chunks(input_stream: IO[AnyStr],
                                 chunk_size: int = 8192) -> IO[bytes]:
        """
        Reads the contents of input_stream and writes it gzip-compressed to
        output_stream in chunks. The chunk size is 8Kb by default, which is a
        standard filesystem block size.
        """
        compressed_stream = io.BytesIO()

        with gzip.open(compressed_stream, 'wb') as gzipped_file:
            chunk_bytes: bytes
            while True:
                chunk = input_stream.read(chunk_size)
                if not chunk:
                    break
                if isinstance(chunk, str):
                    chunk_bytes = chunk.encode('utf-8')
                else:
                    chunk_bytes = chunk

                gzipped_file.write(chunk_bytes)

        # After finishing writing, reset the buffer position so it can be read
        compressed_stream.seek(0)

        return compressed_stream

    @staticmethod
    def _decide_upload_type(
        force_resumable_upload: bool | None,
        content_length: int,
    ) -> UploadType:
        # force resumable
        if force_resumable_upload is True:
            return UploadType.RESUMABLE

        # force simple
        if force_resumable_upload is False:
            return UploadType.SIMPLE

        # decide based on Content-Length
        if content_length > MAX_CONTENT_LENGTH_SIMPLE_UPLOAD:
            return UploadType.RESUMABLE

        return UploadType.SIMPLE

    @staticmethod
    def _split_content_type(content_type: str) -> tuple[str, str | None]:
        content_type_and_encoding_split = content_type.split(';')
        content_type = content_type_and_encoding_split[0].lower().strip()

        encoding = None
        if len(content_type_and_encoding_split) > 1:
            encoding_str = content_type_and_encoding_split[1].lower().strip()
            encoding = encoding_str.split('=')[-1]

        return content_type, encoding

    @staticmethod
    def _format_metadata_key(key: str) -> str:
        """
        Formats the fixed-key metadata keys as wanted by the multipart API.

        Ex: Content-Disposition --> contentDisposition
        """
        parts = key.split('-')
        parts = [parts[0].lower()] + [p.capitalize() for p in parts[1:]]
        return ''.join(parts)

    async def _download(
        self, bucket: str, object_name: str, *,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> bytes:
        # https://cloud.google.com/storage/docs/request-endpoints#encoding
        encoded_object_name = quote(object_name, safe='')
        url = f'{self._api_root_read}/{bucket}/o/{encoded_object_name}'
        headers = headers or {}
        headers.update(await self._headers())

        # aiohttp and requests automatically decompress the body if this
        # argument is not passed, unless a user has explicitly disabled that
        # option at the session level in the case of aiohttp. We follow the
        # user setting by default (by passing None) and only explicitly disable
        # it when a caller explicitly requests a compressed payload.
        auto_decompress = None
        if 'accept-encoding' in {k.lower() for k in headers}:
            auto_decompress = False

        s = AioSession(session) if session else self.session

        data: bytes
        if auto_decompress is False and BUILD_GCLOUD_REST:
            # Requests lib has a different way of reading compressed data. We
            # must pass the stream=True argument and read the response using
            # the 'raw' property.
            response = await s.get(
                url, headers=headers, params=params or {},
                timeout=timeout, stream=True,
            )
            data = response.raw.read()  # type: ignore[attr-defined]
        else:
            response = await s.get(
                url, headers=headers, params=params or {},
                timeout=timeout, auto_decompress=auto_decompress,
            )
            # N.B. the GCS API sometimes returns 'application/octet-stream'
            # when a string was uploaded. To avoid potential weirdness, always
            # return a bytes object.
            try:
                data = await response.read()
            except (AttributeError, TypeError):
                data = response.content  # type: ignore[assignment]

        return data

    async def _download_stream(
        self, bucket: str, object_name: str, *,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> StreamResponse:
        # https://cloud.google.com/storage/docs/request-endpoints#encoding
        encoded_object_name = quote(object_name, safe='')
        url = f'{self._api_root_read}/{bucket}/o/{encoded_object_name}'
        headers = headers or {}
        headers.update(await self._headers())

        # aiohttp and requests automatically decompress the body if this
        # argument is not passed, unless a user has explicitly disabled that
        # option at the session level in the case of aiohttp. We follow the
        # user setting by default (by passing None) and only explicitly disable
        # it when a caller explicitly requests a compressed payload.
        auto_decompress = None
        if 'accept-encoding' in {k.lower() for k in headers}:
            auto_decompress = False

        s = AioSession(session) if session else self.session

        if BUILD_GCLOUD_REST:
            # stream argument is only expected by requests.Session.
            # pylint: disable=unexpected-keyword-arg
            return StreamResponse(
                s.get(
                    url, headers=headers, params=params or {},
                    timeout=timeout, stream=True,
                ),
            )
        return StreamResponse(
            await s.get(
                url, headers=headers, params=params or {},
                timeout=timeout, auto_decompress=auto_decompress,
            ),
        )

    async def _upload_simple(
        self, url: str, object_name: str,
        stream: IO[AnyStr], params: dict[str, str],
        headers: dict[str, str], *,
        session: Session | None = None,
        timeout: int = 30,
    ) -> dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/how-tos/simple-upload
        params['name'] = object_name
        params['uploadType'] = 'media'

        s = AioSession(session) if session else self.session
        resp = await s.post(
            url, data=stream, headers=headers, params=params,
            timeout=timeout,
        )
        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    async def _upload_multipart(
        self, url: str, object_name: str,
        stream: IO[AnyStr], params: dict[str, str],
        headers: dict[str, str],
        metadata: dict[str, Any], *,
        session: Session | None = None,
        timeout: int = 30,
    ) -> dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/how-tos/multipart-upload
        params['uploadType'] = 'multipart'

        metadata_headers = {'Content-Type': 'application/json; charset=UTF-8'}
        metadata = {
            self._format_metadata_key(k): v
            for k, v in metadata.items()
        }
        if 'metadata' in metadata:
            metadata['metadata'] = {
                str(k): str(v) if v is not None else None
                for k, v in metadata['metadata'].items()
            }

        metadata['name'] = object_name

        raw_body: AnyStr = stream.read()
        if isinstance(raw_body, str):
            bytes_body: bytes = raw_body.encode('utf-8')
        else:
            bytes_body = raw_body

        parts = [
            (metadata_headers, json.dumps(metadata).encode('utf-8')),
            ({'Content-Type': headers['Content-Type']}, bytes_body),
        ]
        boundary = choose_boundary()
        body, content_type = encode_multipart_formdata(parts, boundary)
        headers.update({
            'Content-Type': content_type,
            'Content-Length': str(len(body)),
            'Accept': 'application/json',
        })

        s = AioSession(session) if session else self.session
        if not BUILD_GCLOUD_REST:
            # Wrap data in BytesIO to ensure aiohttp does not emit warning
            # when payload size > 1MB
            body = io.BytesIO(body)  # type: ignore[assignment]

        resp = await s.post(
            url, data=body, headers=headers, params=params,
            timeout=timeout,
        )
        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    async def _upload_resumable(
        self, url: str, object_name: str,
        stream: IO[AnyStr], params: dict[str, str],
        headers: dict[str, str], *,
        metadata: dict[str, Any] | None = None,
        session: Session | None = None,
        timeout: int = 30,
    ) -> dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload
        session_uri = await self._initiate_upload(
            url, object_name, params,
            headers, metadata=metadata,
            session=session,
        )
        return await self._do_upload(
            session_uri, stream, headers=headers,
            session=session, timeout=timeout,
        )

    async def _initiate_upload(
        self, url: str, object_name: str,
        params: dict[str, str], headers: dict[str, str],
        *, metadata: dict[str, Any] | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> str:
        params['uploadType'] = 'resumable'

        metadict = (metadata or {}).copy()
        metadict = {
            self._format_metadata_key(k): v
            for k, v in metadict.items()
        }
        if 'metadata' in metadict:
            metadict['metadata'] = {
                str(k): str(v) if v is not None else None
                for k, v in metadict['metadata'].items()
            }

        metadict.update({'name': object_name})
        metadata_ = json.dumps(metadict)

        post_headers = headers.copy()
        post_headers.update({
            'Content-Length': str(len(metadata_)),
            'Content-Type': 'application/json; charset=UTF-8',
            'X-Upload-Content-Type': headers['Content-Type'],
            'X-Upload-Content-Length': headers['Content-Length'],
        })

        s = AioSession(session) if session else self.session
        resp = await s.post(
            url, headers=post_headers, params=params,
            data=metadata_, timeout=timeout,
        )
        session_uri: str = resp.headers['Location']
        return session_uri

    async def _do_upload(
        self, session_uri: str, stream: IO[AnyStr],
        headers: dict[str, str], *, retries: int = 5,
        session: Session | None = None,
        timeout: int = 30,
    ) -> dict[str, Any]:
        s = AioSession(session) if session else self.session

        original_close = stream.close
        original_position = stream.tell()
        # Prevent the stream being closed if put operation fails
        stream.close = lambda: None  # type: ignore[method-assign]
        resp = None
        try:
            for tries in range(retries):
                try:
                    resp = await s.put(
                        session_uri, headers=headers,
                        data=stream, timeout=timeout,
                    )
                except ResponseError:
                    if tries == retries - 1:
                        raise

                    headers.update({'Content-Range': 'bytes */*'})
                    stream.seek(original_position)

                    await sleep(  # type: ignore[func-returns-value]
                        2. ** tries,
                    )
                else:
                    break
        finally:
            original_close()

        if resp is None:
            raise ValueError('null response, this should not happen')

        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    async def patch_metadata(
            self, bucket: str, object_name: str, metadata: dict[str, Any],
            *, params: dict[str, str] | None = None,
            headers: dict[str, str] | None = None,
            session: Session | None = None,
            timeout: int = DEFAULT_TIMEOUT,
    ) -> dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/objects/patch
        encoded_object_name = quote(object_name, safe='')
        url = f'{self._api_root_read}/{bucket}/o/{encoded_object_name}'
        params = params or {}
        headers = headers or {}
        headers.update(await self._headers())
        headers['Content-Type'] = 'application/json'
        body = json.dumps(metadata).encode('utf-8')

        s = AioSession(session) if session else self.session
        resp = await s.patch(
            url, data=body, headers=headers, params=params,
            timeout=timeout,
        )
        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    async def get_bucket_metadata(
        self, bucket: str, *,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        session: Session | None = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> dict[str, Any]:
        url = f'{self._api_root_read}/{bucket}'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        resp = await s.get(
            url, headers=headers, params=params or {},
            timeout=timeout,
        )
        data: dict[str, Any] = await resp.json(content_type=None)
        return data

    async def close(self) -> None:
        await self.session.close()

    async def __aenter__(self) -> 'Storage':
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()



# pylint: disable=line-too-long
"""
This library implements various methods for working with the Google Storage
APIs.

Installation
------------

.. code-block:: console

    $ pip install --upgrade gcloud-aio-storage

Usage
-----

To upload a file, you might do something like the following:

.. code-block:: python

    import aiofiles
    import aiohttp
    from gcloud.aio.storage import Storage


    async with aiohttp.ClientSession() as session:
        client = Storage(session=session)

        async with aiofiles.open('/path/to/my/file', mode="r") as f:
            output = await f.read()
            status = await client.upload(
                'my-bucket-name',
                'path/to/gcs/folder',
                output,
            )
            print(status)

Note that there are multiple ways to accomplish the above, ie,. by making use
of the ``Bucket`` and ``Blob`` convenience classes if that better fits your
use-case.

Of course, the major benefit of using an async library is being able to
parallelize operations like this. Since ``gcloud-aio-storage`` is fully
asyncio-compatible, you can use any of the builtin asyncio method to perform
more complicated operations:

.. code-block:: python

    my_files = {
        '/local/path/to/file.1': 'path/in/gcs.1',
        '/local/path/to/file.2': 'path/in/gcs.2',
        '/local/path/to/file.3': 'different/gcs/path/filename.3',
    }

    async with Storage() as client:
        # Prepare all our upload data
        uploads = []
        for local_name, gcs_name in my_files.items():
            async with aiofiles.open(local_name, mode="r") as f:
                contents = await f.read()
                uploads.append((gcs_name, contents))

        # Simultaneously upload all files
        await asyncio.gather(
            *[
                client.upload('my-bucket-name', path, file_)
                for path, file_ in uploads
            ]
        )

You can also refer to the `smoke test`_ for more info and examples.

Note that you can also let ``gcloud-aio-storage`` do its own session
management, so long as you give us a hint when to close that session:

.. code-block:: python

    async with Storage() as client:
        # closes the client.session on leaving the context manager

    # OR

    client = Storage()
    # do stuff
    await client.close()  # close the session explicitly

File Encodings
--------------

In some cases, ``aiohttp`` needs to transform the objects returned from GCS
into strings, eg. for debug logging and other such issues. The built-in ``await
response.text()`` operation relies on `chardet`_ for guessing the character
encoding in any cases where it can not be determined based on the file
metadata.

Unfortunately, this operation can be extremely slow, especially in cases where
you might be working with particularly large files. If you notice odd latency
issues when reading your results, you may want to set your character encoding
more explicitly within GCS, eg. by ensuring you set the ``contentType`` of the
relevant objects to something suffixed with ``; charset=utf-8``. For example,
in the case of ``contentType='application/x-netcdf'`` files exhibiting latency,
you could instead set ``contentType='application/x-netcdf; charset=utf-8``. See
`Issue #172`_ for more info!

Emulators
---------

For testing purposes, you may want to use ``gcloud-aio-storage`` along with a
local GCS emulator. Setting the ``$STORAGE_EMULATOR_HOST`` environment variable
to the address of your emulator should be enough to do the trick.

For example, using `fsouza/fake-gcs-server`_, you can do:

.. code-block:: console

    docker run -d -p 4443:4443 -v $PWD/my-sample-data:/data fsouza/fake-gcs-server
    export STORAGE_EMULATOR_HOST='http://0.0.0.0:4443'

Any ``gcloud-aio-storage`` requests made with that environment variable set
will query ``fake-gcs-server`` instead of the official GCS API.

Note that some emulation systems require disabling SSL -- if you're using a
custom http session, you may need to disable SSL verification.

Customization
-------------

This library mostly tries to stay agnostic of potential use-cases; as such, we
do not implement any sort of retrying or other policies under the assumption
that we wouldn't get things right for every user's situation.

As such, we recommend configuring your own policies on an as-needed basis. The
`backoff`_ library can make this quite straightforward! For example, you may
find it useful to configure something like:

.. code-block:: python

    class StorageWithBackoff(gcloud.aio.storage.Storage):
        @backoff.on_exception(backoff.expo, aiohttp.ClientResponseError,
                              max_tries=5, jitter=backoff.full_jitter)
        async def copy(self, *args: Any, **kwargs: Any):
            return await super().copy(*args, **kwargs)

        @backoff.on_exception(backoff.expo, aiohttp.ClientResponseError,
                              max_tries=10, jitter=backoff.full_jitter)
        async def download(self, *args: Any, **kwargs: Any):
            return await super().download(*args, **kwargs)

.. _Issue #172: https://github.com/talkiq/gcloud-aio/issues/172
.. _backoff: https://pypi.org/project/backoff/
.. _chardet: https://pypi.org/project/chardet/
.. _fsouza/fake-gcs-server: https://github.com/fsouza/fake-gcs-server
.. _smoke test: https://github.com/talkiq/gcloud-aio/blob/master/storage/tests/integration/smoke_test.py
"""
import importlib.metadata

from .blob import Blob
from .bucket import Bucket
from .storage import SCOPES
from .storage import Storage
from .storage import StreamResponse


__version__ = importlib.metadata.version('gcloud-aio-storage')
__all__ = [
    'Blob',
    'Bucket',
    'SCOPES',
    'Storage',
    'StreamResponse',
    '__version__',
]



import logging
from typing import Any
from typing import TYPE_CHECKING

from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module

from .blob import Blob
from .constants import DEFAULT_TIMEOUT

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import HTTPError as ResponseError
    from requests import Session
else:
    from aiohttp import (  # type: ignore[assignment]
        ClientResponseError as ResponseError,
    )
    from aiohttp import ClientSession as Session  # type: ignore[assignment]

if TYPE_CHECKING:
    from .storage import Storage  # pylint: disable=cyclic-import


log = logging.getLogger(__name__)


class Bucket:
    def __init__(self, storage: 'Storage', name: str) -> None:
        self.storage = storage
        self.name = name

    async def get_blob(
        self, blob_name: str, timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
    ) -> Blob:
        metadata = await self.storage.download_metadata(
            self.name, blob_name,
            timeout=timeout,
            session=session,
        )

        return Blob(self, blob_name, metadata)

    async def blob_exists(
        self, blob_name: str,
        session: Session | None = None,
    ) -> bool:
        try:
            await self.get_blob(blob_name, session=session)
            return True
        except ResponseError as e:
            try:
                if e.status in {404, 410}:  # type: ignore[attr-defined]
                    return False
            except AttributeError:
                if e.code in {404, 410}:  # type: ignore[attr-defined]
                    return False

            raise e

    async def list_blobs(
        self, prefix: str = '', match_glob: str = '',
        delimiter: str = '', session: Session | None = None,
    ) -> list[str]:
        params = {
            'delimiter': delimiter,
            'matchGlob': match_glob,
            'pageToken': '',
            'prefix': prefix,
        }
        items = []
        while True:
            content = await self.storage.list_objects(
                self.name,
                params=params,
                session=session,
            )
            items.extend([x['name'] for x in content.get('items', [])])
            if delimiter:
                items.extend(content.get('prefixes', []))
            params['pageToken'] = content.get('nextPageToken', '')
            if not params['pageToken']:
                break

        return items

    def new_blob(self, blob_name: str) -> Blob:
        return Blob(self, blob_name, {'size': 0})

    async def get_metadata(
            self, params: dict[str, Any] | None = None,
            session: Session | None = None,
    ) -> dict[str, Any]:
        return await self.storage.get_bucket_metadata(
            self.name, params=params,
            session=session,
        )



import binascii
import collections
import datetime
import enum
import hashlib
import io
import os
from typing import Any
from typing import TYPE_CHECKING
from urllib.parse import quote

import rsa
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.auth import decode  # pylint: disable=no-name-in-module
from gcloud.aio.auth import IamClient  # pylint: disable=no-name-in-module
from gcloud.aio.auth import Token  # pylint: disable=no-name-in-module
from pyasn1.codec.der import decoder
from pyasn1_modules import pem
from pyasn1_modules.rfc5208 import PrivateKeyInfo

from .constants import DEFAULT_TIMEOUT

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[assignment]

if TYPE_CHECKING:
    from .bucket import Bucket  # pylint: disable=cyclic-import


HOST = os.environ.get('STORAGE_EMULATOR_HOST', 'storage.googleapis.com')

PKCS1_MARKER = (
    '-----BEGIN RSA PRIVATE KEY-----',
    '-----END RSA PRIVATE KEY-----',
)
PKCS8_MARKER = (
    '-----BEGIN PRIVATE KEY-----',
    '-----END PRIVATE KEY-----',
)
PKCS8_SPEC = PrivateKeyInfo()


class PemKind(enum.Enum):
    """
    Tracks the response of ``pem.readPemBlocksFromFile(key, *args)``>

    Note that the specified method returns ``(marker_id, key_bytes)``, where
    ``marker_id`` is the integer index of the matching ``arg`` (or -1 if no
    match was found.

    For example::

        (marker_id, _) = pem.readPemBlocksFromFile(key, PKCS1_MARKER,
                                                   PCKS8_MARKER)
        if marker_id == -1:
            # "key" did not match either type or was invalid
        if marker_id == 0:
            # "key" matched the zeroth provided marker arg, eg. PKCS1_MARKER
        if marker_id == 1:
            # "key" matched the zeroth provided marker arg, eg. PKCS8_MARKER
    """

    INVALID = -1
    PKCS1 = 0
    PKCS8 = 1


class _SignatureMethod(enum.Enum):
    """
    Indicates where the url signing will be done through Google's
    IAM API or through local signing with a PEM file, which is faster
    but requires that the provided token contains client_email and
    private_key data
    """

    PEM = 0
    IAM_API = 1


class Blob:
    def __init__(
        self, bucket: 'Bucket', name: str,
        metadata: dict[str, Any],
    ) -> None:
        metadata['bucket_name'] = metadata.pop('bucket', '')
        self.__dict__.update(**metadata)

        self.bucket = bucket
        self.name = name
        self.size: int = int(self.size)

    @property
    def chunk_size(self) -> int:
        return self.size + (262144 - (self.size % 262144))

    async def download(
        self, timeout: int = DEFAULT_TIMEOUT,
        session: Session | None = None,
        auto_decompress: bool = True,
    ) -> Any:
        headers = None if auto_decompress else {'accept-encoding': 'gzip'}
        return await self.bucket.storage.download(
            self.bucket.name,
            self.name,
            timeout=timeout,
            session=session,
            headers=headers,
        )

    async def upload(
        self, data: Any,
        content_type: str | None = None,
        session: Session | None = None,
    ) -> dict[str, Any]:
        metadata = await self.bucket.storage.upload(
            self.bucket.name,
            self.name,
            data,
            content_type=content_type,
            session=session,
        )

        metadata['bucket_name'] = metadata.pop('bucket', '')
        self.__dict__.update(metadata)

        return metadata

    async def get_signed_url(  # pylint: disable=too-many-locals
            self, expiration: int, headers: dict[str, str] | None = None,
            query_params: dict[str, Any] | None = None,
            http_method: str = 'GET', iam_client: IamClient | None = None,
            service_account_email: str | None = None,
            token: Token | None = None, session: Session | None = None,
    ) -> str:
        """
        Create a temporary access URL for Storage Blob accessible by anyone
        with the link.

        Adapted from Google Documentation:
        https://cloud.google.com/storage/docs/access-control/signing-urls-manually#python-sample
        """
        if expiration > 604800:
            raise ValueError(
                "expiration time can't be longer than 604800 "
                'seconds (7 days)',
            )

        quoted_name = quote(self.name, safe=b'/~')
        canonical_uri = f'/{self.bucket.name}/{quoted_name}'

        datetime_now = datetime.datetime.now(datetime.timezone.utc)
        request_timestamp = datetime_now.strftime('%Y%m%dT%H%M%SZ')
        datestamp = datetime_now.strftime('%Y%m%d')

        token = token or self.bucket.storage.token
        credential_scope = f'{datestamp}/auto/storage/goog4_request'
        # Try to sign locally if available
        client_email = token.service_data.get('client_email')
        private_key = token.service_data.get('private_key')
        if not client_email or not private_key:
            # Cannot sign locally, so we'll have to use Google's IAM API
            signature_method = _SignatureMethod.IAM_API
            credential = f'{service_account_email}/{credential_scope}'
        else:
            signature_method = _SignatureMethod.PEM
            credential = f'{client_email}/{credential_scope}'

        headers = headers or {}
        headers['host'] = HOST

        ordered_headers = collections.OrderedDict(
            sorted(headers.items(), key=lambda x: x[0].lower()))
        canonical_headers = ''.join(
            f'{str(k).lower()}:{str(v).lower()}\n'
            for k, v in ordered_headers.items()
        )

        signed_headers = ';'.join(
            f'{str(k).lower()}' for k in ordered_headers.keys()
        )

        query_params = query_params or {}
        query_params['X-Goog-Algorithm'] = 'GOOG4-RSA-SHA256'
        query_params['X-Goog-Credential'] = credential
        query_params['X-Goog-Date'] = request_timestamp
        query_params['X-Goog-Expires'] = expiration
        query_params['X-Goog-SignedHeaders'] = signed_headers

        ordered_query_params = collections.OrderedDict(
            sorted(query_params.items()),
        )

        canonical_query_str = '&'.join(
            f'{quote(str(k), safe="")}={quote(str(v), safe="")}'
            for k, v in ordered_query_params.items()
        )

        canonical_req = '\n'.join([
            http_method, canonical_uri,
            canonical_query_str, canonical_headers,
            signed_headers, 'UNSIGNED-PAYLOAD',
        ])
        canonical_req_hash = hashlib.sha256(canonical_req.encode()).hexdigest()

        str_to_sign = '\n'.join([
            'GOOG4-RSA-SHA256', request_timestamp,
            credential_scope, canonical_req_hash,
        ])

        if (signature_method == _SignatureMethod.PEM and private_key
                and isinstance(private_key, str)):
            signed_blob = self.get_pem_signature(str_to_sign, private_key)
        else:
            provided_session: bool = bool(iam_client or session)
            try:
                iam_client = iam_client or IamClient(
                    token=token, session=session)
            except TypeError as e:
                raise TypeError('Blob signing is not yet supported'
                                ' for AUTHORIZED_USER tokens') from e
            signed_blob = await self.get_iam_api_signature(
                str_to_sign,
                iam_client,
                service_account_email,
                session or iam_client.session,  # type: ignore[arg-type]
            )
            if not provided_session:
                await iam_client.close()

        signature = binascii.hexlify(signed_blob).decode()

        return (
            f'https://{HOST}{canonical_uri}?'
            f'{canonical_query_str}&X-Goog-Signature={signature}'
        )

    @staticmethod
    def get_pem_signature(str_to_sign: str, private_key: str) -> bytes:
        # N.B. see the ``PemKind`` enum
        marker_id, key_bytes = pem.readPemBlocksFromFile(
            io.StringIO(private_key), PKCS1_MARKER, PKCS8_MARKER,
        )
        if marker_id == PemKind.INVALID.value:
            raise ValueError('private key is invalid or unsupported')

        if marker_id == PemKind.PKCS8.value:
            # convert from pkcs8 to pkcs1
            key_info, remaining = decoder.decode(
                key_bytes,
                asn1Spec=PKCS8_SPEC,
            )
            if remaining != b'':
                raise ValueError(
                    'could not read PKCS8 key: found extra bytes',
                    remaining,
                )

            private_key_info = key_info.getComponentByName('privateKey')
            key_bytes = private_key_info.asOctets()

        key = rsa.key.PrivateKey.load_pkcs1(key_bytes, format='DER')
        signed_blob = rsa.pkcs1.sign(
            str_to_sign.encode(),
            key,
            'SHA-256',
        )
        return signed_blob

    @staticmethod
    async def get_iam_api_signature(
            str_to_sign: str, iam_client: IamClient,
            service_account_email: str | None, session: Session | None,
    ) -> bytes:
        signed_resp = await iam_client.sign_blob(
            str_to_sign,
            service_account_email=service_account_email,
            session=session,
        )
        return decode(signed_resp['signedBlob'])
